(*
    This file is part of the RI::ML project -
    which includes RenderMan Interface 3.2 bindings
    for Standard ML and a RIB file generator.

    Copyright (C) 2000 ANOQ of the Sun (alias Johnny Andersen).
    
    E-mail: anoq@HardcoreProcessing.com

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    As a special exception, if you do not do anything which is not in
    the spirit of the GNU Library General Public License, you are not
    required to physically compile this software into a separate library,
    since this is generally not possible with current Stanard ML compilers.
    However if you do something which is not in the spirit of the
    GNU Library General Public License you will have to follow the
    licence perpetually - thus disallowing you to use it for any
    commercial purposes at all.

    If you are interested in a warranty or commercial support for this
    software, contact Hardcore Processing <sales@HardcoreProcessing.com>
    for more information.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

functor FRiExample(structure Ri : RENDERMAN_INTERFACE) =
  struct
    fun createFrame number =
          (Ri.FrameBegin number;
             Ri.WorldBegin();
                          (* Notice: The shader "plastic" is a built-in shader
                                     so we should really use RI.PLASTIC here -
                                     but it is just to illustrate the use
                                     of Rt.MakeToken in the ML bindings. *)
               Ri.Surface(Rt.MakeToken "plastic", nil);
               Ri.TransformBegin();
                 Ri.Color (RealArray.fromList [1.0, 1.0, 1.0]);
                 Ri.Translate (0.0, 0.0, 0.0);

                 (* The rotate functions makes an animation based in frame number *)
                 Ri.Rotate (Real.fromInt (number * 4 - 30), 1.0, 0.0, 0.0);
                 Ri.Rotate (Real.fromInt (number * 4 + 130), 0.0, 1.0, 0.0);
                 Ri.Scale (1.5, 1.5, 1.5);
           
                 (* Create an object here... *)
                 Ri.Cone (2.0, 1.0, 360.0, nil);
             
               Ri.TransformEnd();

             Ri.WorldEnd();
           Ri.FrameEnd ();
           ())

    fun test () =
          (Ri.Begin (Rt.MakeToken "output.rib");
             (* Here we place a lightsource at the camera position... *)
             Ri.LightSource(RI.POINTLIGHT, NONE,
                            [(RI.TO, Rt.VPoint {x = 0.0, y = 0.0, z = 0.0}),
                             (RI.INTENSITY, Rt.VFloat 50.0),
                             (RI.LIGHTCOLOR, Rt.VColor (RealArray.fromList [1.0, 1.0, 1.0]))]);
             
             (* Set some display properties *)
             Ri.Format (640, 480, 1.0);
             Ri.Projection (RI.PERSPECTIVE, [(RI.FOV, Rt.VFloat 90.0)]);

             (* Move the camera away from (0, 0, 0) in world coordinates.  *)
             Ri.Translate (0.0, 0.0, 10.0);

             (* Make an animation with 20 frames *)
             List.tabulate (20, createFrame); 
           
           Ri.End ())
  end

(* Now we instantiate the RiExample structure with the RIB
   generator. Here we could also instantiate with other
   RenderMan implementations... *)
structure RiExample = FRiExample(structure Ri = RIB)

val _ = RiExample.test ()
