(*
    This file is part of the RI::ML project -
    which includes RenderMan Interface 3.2 bindings
    for Standard ML and a RIB file generator.

    Copyright (C) 2000 ANOQ of the Sun (alias Johnny Andersen).
    
    E-mail: anoq@HardcoreProcessing.com

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    As a special exception, if you do not do anything which is not in
    the spirit of the GNU Library General Public License, you are not
    required to physically compile this software into a separate library,
    since this is generally not possible with current Stanard ML compilers.
    However if you do something which is not in the spirit of the
    GNU Library General Public License you will have to follow the
    licence perpetually - thus disallowing you to use it for any
    commercial purposes at all.

    If you are interested in a warranty or commercial support for this
    software, contact Hardcore Processing <sales@HardcoreProcessing.com>
    for more information.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

(* All the RI_ prefixes have been removed since they are not
   needed when the functions are grouped in a structure. *)
functor FRenderManInterfaceConsts() =
  struct
    (* Predefined tokens *)

    (* Standard values *)
    val INFINITY = 1.0e38
    val EPSILON  = 1.0e~10

    (* Standard shader variables *)
    val P = Rt.MakeToken "P"
    val PW = Rt.MakeToken "Pw"
    val PZ = Rt.MakeToken "Pz"
    val PREF = Rt.MakeToken "Pref"
    val N = Rt.MakeToken "N"
    val NP = Rt.MakeToken "Np"
    val CS = Rt.MakeToken "Cs"
    val OS = Rt.MakeToken "Os"
    val ST = Rt.MakeToken "st"
    val S = Rt.MakeToken "s"
    val T = Rt.MakeToken "t"
    val KA = Rt.MakeToken "Ka"
    val KD = Rt.MakeToken "Kd"
    val KS = Rt.MakeToken "Ks"
    val KR = Rt.MakeToken "Kr"
    val ROUGHNESS = Rt.MakeToken "roughness"
    val TEXTURENAME = Rt.MakeToken "texturename"
    val SPECULARCOLOR = Rt.MakeToken "specularcolor"

    (* Wrapping modes *)
    val PERIODIC = Rt.MakeToken "periodic"
    val NONPERIODIC = Rt.MakeToken "nonperiodic"

    (* Texture wrapping *)
    (* PERIODIC already defined *)
    val BLACK = Rt.MakeToken "black"
    val CLAMP = Rt.MakeToken "clamp"

    (* Patch types *)
    val BILINEAR = Rt.MakeToken "bilinear"
    val BICUBIC = Rt.MakeToken "bicubic"

    (* Curve types *)
    val LINEAR = Rt.MakeToken "linear"
    val CUBIC = Rt.MakeToken "cubic"

    (* Curve flags *)
    val WIDTH = Rt.MakeToken "width"
    val CONSTANTWIDTH = Rt.MakeToken "constantwidth"

    (* Standard projection shader variables *)
    val FOV = Rt.MakeToken "fov"

    (* Standard light shader variables *)
    val INTENSITY = Rt.MakeToken "intensity"
    val LIGHTCOLOR = Rt.MakeToken "lightcolor"
    val FROM = Rt.MakeToken "from"
    val TO = Rt.MakeToken "to"
    val CONEANGLE = Rt.MakeToken "coneangle"
    val CONEDELTAANGLE = Rt.MakeToken "conedeltaangle"
    val BEAMDISTRIBUTION = Rt.MakeToken "beamdistribution"

    (* Geometric approximation stuff *)
    val FLATNESS = Rt.MakeToken "flatness"

    (* Shading interpolation stuff *)
    val CONSTANT = Rt.MakeToken "constant" (* Also the name of a standard surface shader *)
    val SMOOTH = Rt.MakeToken "smooth"

    (* For Display function *)
    val FILE = Rt.MakeToken "file"
    val FRAMEBUFFER = Rt.MakeToken "framebuffer"
    val RGB = Rt.MakeToken "rgb"
    val RGBA = Rt.MakeToken "rgba"
    val RGBZ = Rt.MakeToken "rgbz"
    val RGBAZ = Rt.MakeToken "rgbaz"
    val A = Rt.MakeToken "a"
    val Z = Rt.MakeToken "z"
    val AZ = Rt.MakeToken "az"

    (* Hider option *)
    val HIDDEN = Rt.MakeToken "hidden"
    val PAINT = Rt.MakeToken "paint"
    val NULLTOKEN = Rt.MakeToken "null"

    (* Standard projection shaders *)
    val PERSPECTIVE = Rt.MakeToken "perspective"
    val ORTHOGRAPHIC = Rt.MakeToken "orthographic"

    (* Standard lightsource shaders *)
    val AMBIENTLIGHT = Rt.MakeToken "ambientlight"
    val DISTANTLIGHT = Rt.MakeToken "distantlight"
    val POINTLIGHT = Rt.MakeToken "pointlight"
    val SPOTLIGHT = Rt.MakeToken "spotlight"

    (* Standard surface shaders *)
    (* CONSTANT is already defined *)
    val MATTE = Rt.MakeToken "matte"
    val METAL = Rt.MakeToken "metal"
    val SHINYMETAL = Rt.MakeToken "shinymetal"
    val PLASTIC = Rt.MakeToken "plastic"
    val PAINTEDPLASTIC = Rt.MakeToken "paintedplastic"

    (* Standard volume shaders *)
    val DEPTHCUE = Rt.MakeToken "depthcue"
    val FOG = Rt.MakeToken "fog"

    (* Standard displacement shaders *)
    val BUMPY = Rt.MakeToken "bumpy"

    (* FIXME: Haven't checked what this is for... *)
    val MINDISTANCE = Rt.MakeToken "mindistance"
    val MAXDISTANCE = Rt.MakeToken "maxdistance"
    val BACKGROUND = Rt.MakeToken "background" (* Also the name of a standard imager shader *)
    val DISTANCE = Rt.MakeToken "distance"
    val AMPLITUDE = Rt.MakeToken "amplitude"

    (* Standard imager shaders *)
    (* BACKGROUND already defined *)

    (* Standard coordinate systems *)
    val RASTER = Rt.MakeToken "raster"
    val SCREEN = Rt.MakeToken "screen"
    val CAMERA = Rt.MakeToken "camera"
    val WORLD  = Rt.MakeToken "world"
    val OBJECT = Rt.MakeToken "object"
    val NDC    = Rt.MakeToken "NDC" (* FIXME: RI_NDC is not mentioned in ri.h in RI Spec. *)

    (* Subdivision surface schemes *)
    val CATMULLCLARK = Rt.MakeToken "catmull-clark"

    (* Subdivision surface tags *)
    val HOLE = Rt.MakeToken "hole"
    val CREASE = Rt.MakeToken "crease"
    val CORNER = Rt.MakeToken "corner"
    val INTERPOLATEBOUNDARY = Rt.MakeToken "interpolateboundary"

    (* Solid operations *)
    val PRIMITIVE = Rt.MakeToken "primitive"
    val INTERSECTION = Rt.MakeToken "intersection"
    val UNION = Rt.MakeToken "union"
    val DIFFERENCE = Rt.MakeToken "difference"

    (* ArchiveRecord types *)
    val COMMENT = Rt.MakeToken "comment"
    val STRUCTURE = Rt.MakeToken "structure"
    val VERBATIM = Rt.MakeToken "verbatim"

    (* Orientations *)
    val INSIDE = Rt.MakeToken "inside"
    val OUTSIDE = Rt.MakeToken "outside"
    val LH = Rt.MakeToken "lh" (* Left hand *)
    val RH = Rt.MakeToken "rh" (* Right hand *)

    (* FIXME: Have not checked what this is for *)
    val IDENTIFIER = Rt.MakeToken "identifier"
    val NAME = Rt.MakeToken "name"
    val SHADINGGROUP = Rt.MakeToken "shadinggroup"
    
    (* Synchronization *)
    val ABORT = Rt.MakeToken "abort" (* Also an error handler type *)

    (* Error handler tokens IGNORE, PRINT, ABORT and HANDLER
       are not used because we have them with the error
       handler functions. *)

    val HERMITESTEP = 2
    val CATMULLROMSTEP = 1
    val BEZIERSTEP = 3
    val BSPLINESTEP = 1
    val POWERSTEP = 4
  end

(* All the RIE_ prefixes have been removed since they are not
   needed when the functions are grouped in a structure. *)
functor FRenderManInterfaceErrorConsts() =
  struct
    (* Error codes *)
    (*  1 - 10  System and file errors
       11 - 20  Program limitations
       21 - 40  State Errors
       41 - 60  Parameter Errors
       61 - 80  Execution Errors *)
    val NOERROR      = 0
    val NOMEM        = 1
    val SYSTEM       = 2
    val NOFILE       = 3
    val BADFILE      = 4
    val VERSION      = 5
    val DISKFULL     = 6
    val INCAPABLE    = 11
    val UNIMPLEMENT  = 12
    val LIMIT        = 13
    val BUG          = 14
    val NOTSTARTED   = 23
    val NESTING      = 24
    val NOTOPTIONS   = 25
    val NOTATTRIBS   = 26
    val NOTPRIMS     = 27
    val ILLSTATE     = 28
    val BADMOTION    = 29
    val BADSOLID     = 30
    val BADTOKEN     = 41
    val RANGE        = 42
    val CONSISTENCY  = 43
    val BADHANDLE    = 44
    val NOSHADER     = 45
    val MISSINGDATA  = 46
    val SYNTAX       = 47
    val MATH         = 61

    (* Error severity levels *)
    val INFO         = 0
    val WARNING      = 1
    val ERROR        = 2
    val SEVERE       = 3
  end

(* The stuff in this functor is actually constants - but they
   have been put here because of their naming conventions -
   i.e. that they are prefixed with Ri rather than RI as for
   the other constants. Thus it belongs in the Ri functor implementing
   the RenderMan Interface, and can easily be added there with
   open RIFCs  *)
functor FRenderManInterfaceFunctionConsts() =
  struct
    (* Builtin error handlers *)
    val ErrorIgnore = (fn _ => (), Rt.MakeToken "ignore") : Rt.ErrorHandler
    val ErrorPrint  = (fn _ => (), Rt.MakeToken "print")  : Rt.ErrorHandler
    val ErrorAbort  = (fn _ => (), Rt.MakeToken "abort")  : Rt.ErrorHandler

    (* Builtin Procedural functions. *)
    val ProcDelayedReadArchive =
          (fn _ => (), Rt.MakeToken "DelayedReadArchive")
          : unit Rt.ProcSubdivFunc

    val ProcRunProgram =
          (fn _ => (), Rt.MakeToken "RunProgram")
          : unit Rt.ProcSubdivFunc

    val ProcDynamicLoad =
          (fn _ => (), Rt.MakeToken "DynamicLoad")
          : unit Rt.ProcSubdivFunc

    val ProcFree =
          (fn _ => (), Rt.MakeToken "ProcFree")
          : unit Rt.ProcFreeFunc

    (* Builtin filter functions *)
    local
      (* Notice: The RenderMan Companion p. 178 says that
                 these functions are never called with
                 abs(x) > xwidth nor abs(y) > ywidth *)
      fun gaussianFilterImpl (x, y, xwidth, ywidth) =
            let
              (* Scale circular filter *)
              val nx = (x * 2.0) / xwidth
              val ny = (y * 2.0) / ywidth
            in
              Math.exp (~2.0 * (nx * nx + ny * ny))
            end

      (* CatmullRom filter is clipped - not widened/narrowed.
         See RenderMan Companion p. 178. *)
      fun catmullRomFilterImpl (x, y, _, _) =
            let
              val r2 = x * x + y * y
              val r = Math.sqrt r2
            in
              if r >= 2.0 then
                0.0
              else if r < 1.0 then
                3.0 * r * r2 - 5.0 * r2 + 2.0
              else
                ~r * r2 + 5.0 * r2 - 8.0 * r + 4.0
            end

      (* Sinc filter is clipped - not widened/narrowed.
         See RenderMan Companion p. 178. *)
      fun sinc1D p =
            if p > ~0.001 andalso p < 0.001 then
              1.0
            else
              (Math.sin p) / p

      fun sincFilterImpl (x, y, _, _) =
            (sinc1D x) * (sinc1D y)

      fun triangle1D (p, width) =
            (* According to the reference implementation
               in the RI Spec. it seems to be OK to assume
               that widths are never 0... And I think
               this implementation should have slightly
               less possibility of giving numerical errors
               than the reference implementation :) *)
            (2.0 - ((Real.abs p) * 2.0)) / width

      fun triangleFilterImpl (x, y, xwidth, ywidth) =
            (triangle1D (x, xwidth)) *
            (triangle1D (y, ywidth))

      fun boxFilterImpl _ =
            1.0
    in
      val GaussianFilter   =
            (gaussianFilterImpl,   Rt.MakeToken "gaussian")
            : Rt.FilterFunc
      
      val CatmullRomFilter =
            (catmullRomFilterImpl, Rt.MakeToken "catmull-rom") (* a.k.a cubic *)
            : Rt.FilterFunc
      
      val SincFilter       =
            (sincFilterImpl,       Rt.MakeToken "sinc")
            : Rt.FilterFunc
      
      val TriangleFilter   =
            (triangleFilterImpl,   Rt.MakeToken "triangle")
            : Rt.FilterFunc
      
      val BoxFilter        =
            (boxFilterImpl,        Rt.MakeToken "box")
            : Rt.FilterFunc
    end

    (* Builting patch basis constants *)
    val HermiteBasis    = (( 2.0,  1.0, ~2.0,  1.0),
                           (~3.0, ~2.0,  3.0, ~1.0),
                           ( 0.0,  1.0,  0.0,  0.0),
                           ( 1.0,  0.0,  0.0,  0.0))
    val CatmullRomBasis = ((~1.0,  3.0, ~3.0,  1.0),
                           ( 2.0, ~5.0,  4.0, ~1.0),
                           (~1.0,  0.0,  1.0,  0.0),
                           ( 0.0, ~2.0,  0.0,  0.0))
    val BezierBasis =     ((~1.0,  3.0, ~3.0,  1.0),
                           ( 3.0, ~6.0,  3.0,  0.0),
                           (~3.0,  3.0,  0.0,  0.0),
                           ( 1.0,  0.0,  0.0,  0.0))
    val BSplineBasis =    ((~1.0,  3.0, ~3.0,  1.0),
                           ( 3.0, ~6.0,  3.0,  0.0),
                           (~3.0,  0.0,  3.0,  0.0),
                           ( 1.0,  4.0,  1.0,  0.0))
    val PowerBasis =      (( 1.0,  0.0,  0.0,  0.0),
                           ( 0.0,  1.0,  0.0,  0.0),
                           ( 0.0,  0.0,  1.0,  0.0),
                           ( 0.0,  0.0,  0.0,  1.0))
  end

structure RI = FRenderManInterfaceConsts()
structure RIE = FRenderManInterfaceErrorConsts()
structure RIFCs = FRenderManInterfaceFunctionConsts()