(*
    This file is part of the RI::ML project -
    which includes RenderMan Interface 3.2 bindings
    for Standard ML and a RIB file generator.

    Copyright (C) 2000 ANOQ of the Sun (alias Johnny Andersen).
    
    E-mail: anoq@HardcoreProcessing.com

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    As a special exception, if you do not do anything which is not in
    the spirit of the GNU Library General Public License, you are not
    required to physically compile this software into a separate library,
    since this is generally not possible with current Stanard ML compilers.
    However if you do something which is not in the spirit of the
    GNU Library General Public License you will have to follow the
    licence perpetually - thus disallowing you to use it for any
    commercial purposes at all.

    If you are interested in a warranty or commercial support for this
    software, contact Hardcore Processing <sales@HardcoreProcessing.com>
    for more information.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

(* This is based on "The RenderMan(tm) Companion" by
     Steve Upstill,
   "Advanced RenderMan: Creating CGI for Motion Pictures" by
     Anthony A. Apodaca & Larry Gritz and
   "The RenderMan Interface Specification version 3.2" from
     Pixar. *)
(* All the Ri prefixes have been removed since they are not
   needed when the functions are grouped in a structure. *)
signature RENDERMAN_INTERFACE =
  sig
    (* RenderMan functions *)
    (* Notice: We don't need the vectorized versions of
       the functions in ML do we?
       Even if we want to call it from, say, Pascal? *)
    
    (* General *)
    val ErrorHandler : Rt.ErrorHandler -> unit
    val GetLastError : unit -> int
    val Declare : string * string -> Rt.Token
               (* FIXME: I assume that Synchronize takes an Rt.Token *)
    val Synchronize : Rt.Token -> unit (* In Adv. RenderMan, not in Companion or RI Spec.. *)
    val ArchiveRecord : Rt.Token * string * (Rt.Value list) -> unit (* In Adv. RenderMan and RI Spec., not in Companion. *)
    val ReadArchive : Rt.Token * Rt.ArchiveCallback -> unit (* In Adv. RenderMan and RI Spec., not in Companion. *)

    (* Rendering Contexts *)
    val GetContext : unit -> Rt.ContextHandle
    val Context : Rt.ContextHandle -> unit

    (* Begin-end block *)
    val Begin : Rt.Token -> unit (* The Rt.Token parameter seems to be new in RI Spec. 3.2 *)
    val End : unit -> unit

    (* Frame block *)
    val FrameBegin : int -> unit
    val FrameEnd : unit -> unit

    (* World block *)
    val WorldBegin : unit -> unit
    val WorldEnd : unit -> unit

    (* Attribute block *)
    val AttributeBegin : unit -> unit
    val AttributeEnd : unit -> unit

    (* Transform block *)
    val TransformBegin : unit -> unit
    val TransformEnd : unit -> unit
    
    (* Solid block *)
    val SolidBegin : Rt.Token -> unit
    val SolidEnd : unit -> unit

    (* Object block *)
    (* The Rt.ObjectHandle option parameter is to allow
       reuse of existing object handles *)
    val ObjectBegin : Rt.ObjectHandle option -> Rt.ObjectHandle
    val ObjectEnd : unit -> unit

    (* Motion block *)
    val MotionBegin : real list -> unit
    val MotionEnd : unit -> unit
    
    (* Options *)
    val Format : int * int * real -> unit
    val FrameAspectRatio : real -> unit
    val ScreenWindow : real * real * real * real -> unit
    val CropWindow : real * real * real * real -> unit
    val Projection : Rt.Token * Rt.Parameterlist -> unit (* Name is Rt.Token in RI Spec. 3.2 *)
    val Clipping : real * real -> unit
    val ClippingPlane : real * real * real * real * real * real -> unit (* New in RI Spec. 3.2 *)
    val DepthOfField : real * real * real -> unit
    val Shutter : real * real -> unit
    val PixelVariance : real -> unit
    val PixelSamples : real * real -> unit
    val PixelFilter : Rt.FilterFunc * real * real -> unit
    val Exposure : real * real -> unit
    val Imager : Rt.Token * Rt.Parameterlist -> unit (* Name is Rt.Token in RI Spec. 3.2 *)
    val Quantize : Rt.Token * int * int * int * real -> unit
    val Display : Rt.Token * Rt.Token * Rt.Token * Rt.Parameterlist -> unit (* Name is Rt.Token in RI Spec. 3.2 *)
    val Hider : Rt.Token * Rt.Parameterlist -> unit
                    (* N, nRGB (Nx3 matrix), RGBn (3xN matrix) *)
    val ColorSamples : int *
                       ({r : real, g : real, b : real} list) *
                       {r : real list, g : real list, b : real list} -> unit
    val RelativeDetail : real -> unit
    val Option : Rt.Token * Rt.Parameterlist -> unit (* Name is Rt.Token in RI Spec. 3.2 *)

    (* Lights *)
    (* The Rt.LightHandle option parameters are for allowing
       reuse of existing light handles *)
    val LightSource : Rt.Token * (Rt.LightHandle option) * Rt.Parameterlist -> Rt.LightHandle (* Shadername is Rt.Token in RI Spec. 3.2 *)
    val AreaLightSource : Rt.Token * (Rt.LightHandle option) * Rt.Parameterlist -> Rt.LightHandle (* Shadername is Rt.Token in RI Spec. 3.2 *)

    (* Attributes *)
    val Attribute : Rt.Token * Rt.Parameterlist -> unit (* Name is Rt.Token in RI Spec. 3.2 *)
    val Color : Rt.Color -> unit
    val Opacity : Rt.Color -> unit
    val Surface : Rt.Token * Rt.Parameterlist -> unit (* Shadername is Rt.Token in RI Spec. 3.2 *)
    val Atmosphere : Rt.Token * Rt.Parameterlist -> unit (* Shadername is Rt.Token in RI Spec. 3.2 *)
    val Interior : Rt.Token * Rt.Parameterlist -> unit (* Shadername is Rt.Token in RI Spec. 3.2 *)
    val Exterior : Rt.Token * Rt.Parameterlist -> unit (* Shadername is Rt.Token in RI Spec. 3.2 *)
    val Displacement : Rt.Token * Rt.Parameterlist -> unit
    val Illuminate : Rt.LightHandle * bool -> unit
    (* Deformation and Transformation shaders have been removed in RI Spec. 3.2 *)
    val TextureCoordinates : real * real * real * real *
                             real * real * real * real -> unit
    val ShadingRate : real -> unit
    val ShadingInterpolation : Rt.Token -> unit
    val Matte : bool -> unit
    val Bound : Rt.Bound -> unit
    val Detail : Rt.Bound -> unit
    val DetailRange : real * real * real * real -> unit
    val GeometricApproximation : Rt.Token * real -> unit
    val Orientation : Rt.Token -> unit
    val ReverseOrientation : unit -> unit
    val Sides : int -> unit
    val Basis : Rt.Basis * int * Rt.Basis * int -> unit
    
    (* Transformations *)
    val Identity : unit -> unit
    val Transform : Rt.Matrix -> unit
    val ConcatTransform : Rt.Matrix -> unit
    val Perspective : real -> unit
    val Translate : real * real * real -> unit
    val Rotate : real * real * real * real -> unit
    val Scale : real * real * real -> unit
    val Skew : real * real * real * real * real * real * real -> unit
    val CoordinateSystem : Rt.Token -> unit
    val CoordSysTransform : Rt.Token -> unit (* In Adv. RenderMan, not in Companion. *)
    val TransformPoints : Rt.Token * Rt.Token * (Rt.Point list) -> Rt.Point list

    (* Polygons *)
    val Polygon : Rt.Parameterlist -> unit
    val GeneralPolygon : (int list) * Rt.Parameterlist -> unit
    val PointsPolygons : (int list) * (int list) * Rt.Parameterlist -> unit
    val PointsGeneralPolygons : (int list) * (int list) * (int list) *
                                Rt.Parameterlist -> unit

    (* Experimental polygon functions for increased efficiency in ML. *)
                                (* list of polygons *)
    (* val PointsPolygonsRefList : (Rt.Point ref list list) *
                                (* list of points
                                   (FIXME: should be in Rt.ParameterList
                                   as token "P" *)
                                   Rt.Point ref list *
                                   Rt.Parameterlist -> unit
    *)
    (* val PointsPolygonsList : Rt.Point list list -> unit *)
    (* End of experimental functions *)

    (* Patches *)
    val Patch : Rt.Token * Rt.Parameterlist -> unit
    val PatchMesh : Rt.Token * int * Rt.Token *
                               int * Rt.Token *
                               Rt.Parameterlist -> unit
    (* I think they just forgot parameterlist in NuPatch in
       Adv. RenderMan because it's in Companion. *)
    val NuPatch : int * int * (real list) * real * real *
                  int * int * (real list) * real * real *
                  Rt.Parameterlist -> unit

    (* Notice: This function is listed under "attributes"
               in RenderMan Companion, but I think it
               really belongs here... *)
    val TrimCurve : (int list) * (int list) *
                    (real list) * (real list) * (real list) *
                    (int list) * (real list) * (real list) *
                    (real list) -> unit
    
    (* SubdivisionMesh *)
    val SubdivisionMesh : Rt.Token * (int list) * (int list) * (Rt.Token list) *
                          (int list) * (int list) * (real list) *
                          Rt.Parameterlist -> unit (* Scheme is Rt.Token in RI Spec. 3.2 *)

    (* Lightweight primitives *)
    val Points : Rt.Parameterlist -> unit
    val Curves : Rt.Token * (int list) * Rt.Token * Rt.Parameterlist -> unit
    val Blobby : int * (int list) * (real list) * (string list) *
                 Rt.Parameterlist -> unit

    (* Quadrics *)
    val Sphere : real * real * real * real * Rt.Parameterlist -> unit
    val Cone : real * real * real * Rt.Parameterlist -> unit
    val Cylinder : real * real * real * real * Rt.Parameterlist -> unit
    val Hyperboloid : Rt.Point * Rt.Point * real * Rt.Parameterlist -> unit
    val Paraboloid : real * real * real * real * Rt.Parameterlist -> unit
    val Disk : real * real * real * Rt.Parameterlist -> unit
    val Torus : real * real * real * real * real * Rt.Parameterlist -> unit

    (* General objects *)
    val ObjectInstance : Rt.ObjectHandle -> unit
    val Procedural : ('a Rt.Blind) * Rt.Bound *
                     ('a Rt.ProcSubdivFunc) *
                     ('a Rt.ProcFreeFunc) -> unit
    val Geometry : Rt.Token * Rt.Parameterlist -> unit (* Only mentioned in RI Spec. for defining implementation specific geometric primitives *)

    (* Map making *)    
    val MakeTexture : string * string * Rt.Token * Rt.Token *
                      Rt.FilterFunc * real * real *
                      Rt.Parameterlist -> unit
    val MakeLatLongEnvironment : string * string * Rt.FilterFunc *
                                 real * real * Rt.Parameterlist -> unit
    val MakeCubeFaceEnvironment : string * string * string *
                                  string * string * string *
                                  string * real * Rt.FilterFunc *
                                  real * real * Rt.Parameterlist -> unit
    val MakeShadow : string * string * Rt.Parameterlist -> unit
    (* MakeBump has been removed in RI Spec. 3.2 *)


    (* The rest is actually constants, but they are prefixed
       with Ri in the RenderMan Interface and contains mixed case letters -
       so they belong here. They have been implemented in a separate
       module (RIFCs) so that RenderMan implementors don't
       have to reimplement it. *)
    (* Builtin error handlers *)
    val ErrorIgnore : Rt.ErrorHandler
    val ErrorPrint  : Rt.ErrorHandler
    val ErrorAbort  : Rt.ErrorHandler

    (* Builtin Procedural functions. *)
    val ProcDelayedReadArchive : unit Rt.ProcSubdivFunc
    val ProcRunProgram         : unit Rt.ProcSubdivFunc
    val ProcDynamicLoad        : unit Rt.ProcSubdivFunc
    val ProcFree               : unit Rt.ProcFreeFunc

    (* Builtin filter functions *)
    val GaussianFilter   : Rt.FilterFunc
    val CatmullRomFilter : Rt.FilterFunc
    val SincFilter       : Rt.FilterFunc
    val TriangleFilter   : Rt.FilterFunc
    val BoxFilter        : Rt.FilterFunc

    (* Builting patch basis constants *)
    val HermiteBasis    : Rt.Basis
    val CatmullRomBasis : Rt.Basis
    val BezierBasis     : Rt.Basis
    val BSplineBasis    : Rt.Basis
    val PowerBasis      : Rt.Basis
  end